# -*- coding: utf-8 -*-

from . import importing, validation
from .version import __version__, __url__, __author__, __email__, __description__


# ensure ffmpeg binaries needed by various submodules are on path
import ffmpeg as _ffmpeg
_ffmpeg.add_to_path()


import threading
import time
import numpy as np
from imgui_bundle import imgui, implot, immapp, hello_imgui, glfw_utils
import glfw

from .. import gaze_headref
from . import utils

class TargetPos:
    def __init__(self, timestamp: float, frame_idx: int, cam_pos: np.ndarray):
        self.timestamp = timestamp
        self.frame_idx = frame_idx
        self.cam_pos   = cam_pos

# GUI for synchronizing two signals
class GUI:
    def __init__(self, use_thread=True):
        self._should_exit = False
        self._should_init = True
        self._use_thread = use_thread # NB: on MacOSX the GUI needs to be on the main thread, see https://github.com/pthom/hello_imgui/issues/33
        self._thread = None

        self.title      = ''
        self.ival       = -1
        self._data_lock : threading.Lock = threading.Lock()
        self.gaze_data  : dict[str, np.ndarray] = {}    # ['ts', 'x', 'y']
        self.target_data: dict[str, np.ndarray] = {}
        self.offset_t   : float = 0.
        self.offset_t_ori : float = 0.
        self._offset_xy : list[float] = [0., 0.]
        self._dragging  : list[bool] = [False, False]
        self._temp_off  = np.zeros((3,))
        self._should_rescale = True
        self.is_started : bool = False
        self.is_done    : bool = None

        # cache values
        self._gaze_data_offset  : dict[str, np.ndarray] = {}    # ['ts', 'x', 'y']
        self._gaze_data_plot_pix: dict[str, np.ndarray] = {'x': None, 'y': None}
        self._last_mouse_pos    : dict[int,tuple[float,float]] = {0: None, 1: None}
        self._hovered           : int = None
        self._held              : int = None
        self._drag_origin       : implot.Point = None

        self._window_flags = int(
                                    imgui.WindowFlags_.no_title_bar |
                                    imgui.WindowFlags_.no_collapse |
                                    imgui.WindowFlags_.no_scrollbar |
                                    imgui.WindowFlags_.no_scroll_with_mouse
                                )

    def __del__(self):
        self.stop()

    def start(self):
        if self._use_thread:
            if self._thread is not None:
                raise RuntimeError('The gui is already running, cannot start again')
            self._thread = threading.Thread(target=self._thread_start_fun)
            self._thread.start()
        else:
            self._thread_start_fun()

    def _thread_start_fun(self):
        self._should_exit = False
        self._user_closed_window = False

        def close_callback(window: glfw._GLFWwindow):
            self._user_closed_window = True

        def post_init():
            imgui.get_io().config_viewports_no_decoration = False
            imgui.get_io().config_viewports_no_auto_merge = True

            glfw.set_window_close_callback(glfw_utils.glfw_window_hello_imgui(), close_callback)

        params = hello_imgui.RunnerParams()
        params.app_window_params.restore_previous_geometry = False
        params.ini_folder_type = hello_imgui.IniFolderType.temp_folder  # so we don't have endless ini files in the app folder, since we don't use them anyway (see previous line, restore_previous_geometry = False)
        params.app_window_params.hidden = True
        params.fps_idling.fps_idle = 0
        params.callbacks.show_gui  = self._gui_func
        params.callbacks.post_init = post_init

        addons = immapp.AddOnsParams()
        addons.with_implot = True
        immapp.run(params, addons)

    def get_state(self):
        return (self._user_closed_window, self.is_done)

    def set_data(self, title: str, ival: int, gazes: dict[int,gaze_headref.Gaze], target_positions: dict[int,TargetPos],
                 offset_t = 0.):
        if not gazes or not target_positions:
            raise ValueError('No gaze or target position data was provided')

        new_title = f'{title} (episode {ival})'
        # the next can only be called once the gui is actually started. Avoid race condition
        while not self.is_started:
            time.sleep(0.01)
        # this is just for show, doesn't trigger an update. But lets keep them in sync
        hello_imgui.get_runner_params().app_window_params.window_title = new_title
        # actually update window title
        win = glfw_utils.glfw_window_hello_imgui()
        glfw.set_window_title(win, new_title)
        hello_imgui.get_runner_params().app_window_params.hidden = False
        with self._data_lock:
            self.offset_t   = offset_t
            self.offset_t_ori = offset_t
            self._offset_xy = [0., 0.]
            self.is_done    = False
            self._should_rescale = True

            self.gaze_data['ts']= np.array([s.timestamp       for fr in gazes for s in gazes[fr]],'float')/1000 # ms -> s
            self.gaze_data['x'] = np.array([s.gaze_pos_vid[0] for fr in gazes for s in gazes[fr]],'float')
            self.gaze_data['y'] = np.array([s.gaze_pos_vid[1] for fr in gazes for s in gazes[fr]],'float')

            self.target_data['ts']  = np.array([target_positions[fr].timestamp  for fr in target_positions],'float')/1000 # ms -> s
            self.target_data['x']   = np.array([target_positions[fr].cam_pos[0] for fr in target_positions],'float')
            self.target_data['y']   = np.array([target_positions[fr].cam_pos[1] for fr in target_positions],'float')

            # set begin to 0
            t0 = min([self.gaze_data['ts'][0] , self.target_data['ts'][0] ])
            self.gaze_data  ['ts'] -= t0
            self.target_data['ts'] -= t0

            # set up cache
            self._gaze_data_offset['ts']    = self.gaze_data['ts']+self.offset_t
            self._gaze_data_offset['x']     = self.gaze_data['x'].copy()
            self._gaze_data_offset['y']     = self.gaze_data['y'].copy()

    def _gui_func(self):
        self.is_started = True

        # check if we should exit
        if self._should_exit:
            # and kill
            hello_imgui.get_runner_params().app_shall_exit = True
            # nothing more to do
            return

        # check we have data to plot
        if not self.gaze_data:
            return

        # check we should do some initializing (do this now, can't be done in post_init during the first frame)
        if self._should_init:
            win     = glfw_utils.glfw_window_hello_imgui()
            w_bounds= utils.get_current_monitor(*glfw.get_window_pos(win))[1]
            w_bounds= utils.adjust_bounds_for_framesize(w_bounds, glfw.get_window_frame_size(win))
            glfw.set_window_pos (win, *w_bounds.position)
            glfw.set_window_size(win, *w_bounds.size)
            self._should_init = False

        if disabled:=self.offset_t==0.:
            imgui.begin_disabled()
        if imgui.button('Reset to 0 ms'):
            self.offset_t = 0.
            self._update_cache(None)
        if disabled:
            imgui.end_disabled()
        imgui.same_line()
        if self.offset_t_ori!=0.:
            if disabled:=self.offset_t==self.offset_t_ori:
                imgui.begin_disabled()
            if imgui.button(f'Reset to {self.offset_t_ori*1000:.1f} ms'):
                self.offset_t = self.offset_t_ori
                self._update_cache(None)
            if disabled:
                imgui.end_disabled()
            imgui.same_line()
        if imgui.button('Done', (-1,0)):
            self.is_done = True

        with self._data_lock:
            # handle gaze moving with arrow keys
            move_dir = -1 if imgui.is_key_pressed(imgui.Key.left_arrow) else 1 if imgui.is_key_pressed(imgui.Key.right_arrow) else 0
            if move_dir:
                step = 1.   if imgui.is_key_down(imgui.Key.mod_ctrl)  else \
                       0.15 if imgui.is_key_down(imgui.Key.mod_shift) else \
                       0.01 if imgui.is_key_down(imgui.Key.mod_alt)   else \
                       0.05
                self.offset_t += move_dir*step
                self._update_cache(None)

            gt = self._gaze_data_offset['ts']
            gx = self._gaze_data_offset['x']
            gy = self._gaze_data_offset['y']
            if self._held is not None:
                gt = gt+self._temp_off[0]
                gx = gx+self._temp_off[1]
                gy = gy+self._temp_off[2]

            if implot.begin_subplots('##xy_plots',2,1,(-1,-1),implot.SubplotFlags_.link_all_x):
                for d in range(2):
                    if self._should_rescale:
                        implot.set_next_axes_to_fit()
                        if d==1:
                            self._should_rescale = False
                    if implot.begin_plot('##X',flags=implot.Flags_.no_mouse_text):
                        implot.setup_axis(implot.ImAxis_.x1, None if d==0 else 'time (s)')
                        implot.setup_axis(implot.ImAxis_.y1, 'horizontal coordinate (pix)' if d==0 else 'vertical coordinate (pix)')
                        if self._hovered==0:
                            implot.push_style_var(implot.StyleVar_.line_weight, implot.get_style().line_weight*2)
                            imgui.set_mouse_cursor(imgui.MouseCursor_.hand)
                        implot.plot_line("gaze", gt, gx if d==0 else gy)
                        if self._hovered==0:
                            implot.pop_style_var()
                        implot.plot_line("target", self.target_data['ts'], self.target_data['x' if d==0 else 'y'])
                        if implot.is_plot_hovered():
                            self._do_data_drag(0)
                        if d==0:
                            # position annotation outside axes, clamping will put it in the corner
                            ax_lims = implot.get_plot_limits()
                            implot.annotation(ax_lims.x.max,ax_lims.y.min,implot.get_style().color_(implot.Col_.inlay_text),(0,0),True,f'offset: {(self.offset_t+self._temp_off[0])*1000:.1f}ms')
                        implot.end_plot()

                implot.end_subplots()

                ws = imgui.get_window_size()
                ts = imgui.calc_text_size('(?)')
                fp = imgui.get_style().frame_padding
                imgui.set_cursor_pos((ws.x-ts.x-fp.x, ws.y+imgui.get_scroll_y()-ts.y-fp.y))
                imgui.text('(?)')
                if imgui.is_item_hovered():
                    imgui.begin_tooltip()
                    imgui.push_text_wrap_pos(min(imgui.get_font_size() * 35, ws.x))
                    imgui.text_unformatted(
                        'Drag the gaze signal to align the two signals in time with each other. '
                        'You can also shift the gaze signal in time by using the arrow keys (hold shift '
                        'for larger steps, control for extra large steps and alt for smaller). '
                        'The horizontal offset is the applied time shift (indicated by the value in the lower-right '
                        'corner of the upper plot). Any vertical shift is not stored, but can be useful when aligning the two signals. '
                        'When done aligning the two signals, press done atop the window.'
                        )
                    imgui.pop_text_wrap_pos()
                    imgui.end_tooltip()

    def _do_data_drag(self, ax_idx):
        gid = imgui.get_id(f"##gaze_data_drag_{ax_idx}")
        imgui.push_id(gid)
        p = implot.get_plot_mouse_pos()
        if self._last_mouse_pos[ax_idx] is None or self._last_mouse_pos[ax_idx].x!=p.x:
            self._last_mouse_pos[ax_idx] = p
            distance = self._distance_to_gaze_signal(p, ax_idx)
            if distance < 4.*hello_imgui.dpi_window_size_factor():
                # hovering
                self._hovered = ax_idx
            elif self._hovered==ax_idx:
                self._hovered = None
        released = False
        drag_offset = imgui.ImVec2()
        if self._hovered==ax_idx and imgui.is_mouse_clicked(imgui.MouseButton_.left):
            self._held = ax_idx
            self._drag_origin = p
            imgui.internal.set_active_id(gid,imgui.internal.get_current_window())
        elif self._held==ax_idx:
            if imgui.is_mouse_down(imgui.MouseButton_.left):
                imgui.internal.set_active_id(gid,imgui.internal.get_current_window())
                drag_offset = imgui.ImVec2(p.x-self._drag_origin.x, p.y-self._drag_origin.y)
            else:
                self._held = None
                self._drag_origin = None
                released = True

        # process drag, apply
        if self._held==ax_idx:
            self._temp_off[0]        = drag_offset.x
            self._temp_off[ax_idx+1] = drag_offset.y
        elif released:
            self.offset_t           += self._temp_off[0]
            self._offset_xy[ax_idx] += self._temp_off[ax_idx+1]
            self._temp_off          = np.zeros((3,))
            self._update_cache(ax_idx)
        imgui.pop_id()

    def _update_cache(self, ax_idx: int=None):
        self._gaze_data_offset['ts'] = self.gaze_data['ts'] + self.offset_t
        if ax_idx in [0, 2]:
            self._gaze_data_offset['x'] = self.gaze_data['x'] + self._offset_xy[0]
        elif ax_idx in [1, 2]:
            self._gaze_data_offset['y'] = self.gaze_data['y'] + self._offset_xy[1]
        self._gaze_data_plot_pix = {'x': None, 'y': None}

    def _distance_to_gaze_signal(self, pos: implot.Point, ax_idx: int):
        ax = 'x' if ax_idx==0 else 'y'
        if self._gaze_data_plot_pix[ax] is None:
            self._gaze_data_plot_pix[ax] = np.array([(p.x,p.y) for p in (implot.plot_to_pixels(implot.Point(x,y)) for x,y in zip(self._gaze_data_offset['ts'], self._gaze_data_offset[ax]))])

        pos = implot.plot_to_pixels(pos)
        return np.min(np.hypot(self._gaze_data_plot_pix[ax][:,0]-pos.x, self._gaze_data_plot_pix[ax][:,1]-pos.y))

    def stop(self):
        self._should_exit = True
        if self._thread is not None:
            self._thread.join()
        self._thread = None


import cv2
import numpy as np
import pathlib
import importlib.resources
import shutil
import math
import typing
from matplotlib import colors

from .... import aruco, drawing, marker, plane, transforms

def deploy_maker(output_dir: str|pathlib.Path):
    output_dir = pathlib.Path(output_dir)
    if not output_dir.is_dir():
        raise RuntimeError('the requested directory "%s" does not exist' % output_dir)

    # copy over all files
    for r in ['poster.tex']:
        with importlib.resources.path(__package__, r) as p:
            shutil.copyfile(p, str(output_dir/r))

    deploy_marker_images(output_dir)

def deploy_marker_images(output_dir: str|pathlib.Path):
    from .. import get_validation_setup

    output_dir = pathlib.Path(output_dir) / "all-markers"
    if not output_dir.is_dir():
        output_dir.mkdir()

    # get validation setup
    validationSetup = get_validation_setup()

    # generate and store the markers
    aruco.deploy_marker_images(output_dir, 1000, ValidationPlane.default_aruco_dict, validationSetup['markerBorderBits'])

def deploy_default_pdf(output_file_or_dir: str|pathlib.Path):
    output_file_or_dir = pathlib.Path(output_file_or_dir)
    if output_file_or_dir.is_dir():
        output_file_or_dir = output_file_or_dir / 'poster.pdf'

    with importlib.resources.path(__package__,'poster.pdf') as p:
        shutil.copyfile(p, str(output_file_or_dir))


class ValidationPlane(plane.Plane):
    poster_image_filename = 'referencePoster.png'
    default_aruco_dict    = cv2.aruco.DICT_4X4_250

    def __init__(self, config_dir: str|pathlib.Path|None, validation_config: dict[str,typing.Any]=None, **kwarg):
        from .. import get_markers, get_validation_setup
        # NB: if config_dir is None, the default config will be used

        if config_dir is not None:
            config_dir = pathlib.Path(config_dir)

        # get validation config, if needed
        if validation_config is None:
            validation_config = get_validation_setup(config_dir)
        self.config = validation_config

        # get marker width
        if self.config['mode'] == 'deg':
            self.cell_size_mm = 2.*math.tan(math.radians(.5))*self.config['distance']*10
        else:
            self.cell_size_mm = 10 # 1cm
        markerSize = self.cell_size_mm*self.config['markerSide']

        # get board size
        plane_size = plane.Coordinate(self.config['gridCols']*self.cell_size_mm, self.config['gridRows']*self.cell_size_mm)

        # get targets first, so that they can be drawn on the reference image
        self.targets: dict[int,marker.Marker] = {}
        origin = self._get_targets(config_dir, self.config)

        # call base class
        markers = get_markers(config_dir, self.config['markerPosFile'])
        ref_image_store_path = None
        if 'ref_image_store_path' in kwarg:
            ref_image_store_path = kwarg.pop('ref_image_store_path')
        elif config_dir is not None:
            ref_image_store_path = config_dir / self.poster_image_filename
        super(ValidationPlane, self).__init__(markers, markerSize, plane_size, ValidationPlane.default_aruco_dict, self.config['markerBorderBits'],self.cell_size_mm, "mm", ref_image_store_path=ref_image_store_path, ref_image_size=self.config['referencePosterSize'],**kwarg)

        # set center
        self.set_origin(origin)

    def set_origin(self, origin: plane.Coordinate):
        # set origin of plane. Origin location is on current (not original) plane
        # so set_origin([5., 0.]) three times in a row shifts the origin rightward by 15 units
        for i in self.targets:
            self.targets[i].shift(-np.array(origin))
        super(ValidationPlane, self).set_origin(origin)

    def _get_targets(self, config_dir, validationSetup) -> plane.Coordinate:
        """ poster space: (0,0) is origin (might be center target), (-,-) bottom left """
        from .. import get_targets

        # read in target positions
        targets = get_targets(config_dir, validationSetup['targetPosFile'])
        if targets is not None:
            targets['center'] = list(targets[['x','y']].values)
            targets['center'] *= self.cell_size_mm
            targets = targets.drop(['x','y'], axis=1)
            self.targets = {idx:marker.Marker(idx,**kwargs) for idx,kwargs in zip(targets.index.values,targets.to_dict(orient='records'))}
            origin = plane.Coordinate(*targets.loc[validationSetup['centerTarget']].center.copy())  # NB: need origin in scaled space
        else:
            origin = plane.Coordinate(0.,0.)
        return origin

    def _store_reference_image(self, path: pathlib.Path, width: int) -> np.ndarray:
        # first call superclass method to generate image without targets
        img = super(ValidationPlane, self)._store_reference_image(path, width)
        height = img.shape[0]

        # add targets
        subPixelFac = 8   # for sub-pixel positioning
        for key in self.targets:
            # 1. determine position on image
            circlePos = transforms.to_image_pos(*self.targets[key].center, self.bbox,[width,height])

            # 2. draw
            clr = tuple([int(i*255) for i in colors.to_rgb(self.targets[key].color)[::-1]])  # need BGR color ordering
            drawing.openCVCircle(img, circlePos, 15, clr, -1, subPixelFac)

        if path:
            cv2.imwrite(path, img)

        return img




import pathlib
import pandas as pd

from . import DataQualityType

def collect_data_quality(rec_dirs: list[str | pathlib.Path], file_name: str|dict[str,str]='dataQuality.tsv', col_for_parent=None):
    # 1. collect all data quality metrics from the provided directories
    rec_files: list = []
    idx_vals = ['recording']
    if isinstance(file_name,dict):
        for f in file_name:
            for d in rec_dirs:
                f_path = pathlib.Path(d)/file_name[f]
                if not f_path.is_file():
                    continue
                kwargs = {'recording': f_path.parent.name, 'plane': f}
                if col_for_parent:
                    kwargs[col_for_parent] = f_path.parent.parent.name
                rec_files.append((f_path,kwargs))
        idx_vals.append('plane')
        if col_for_parent:
            idx_vals.insert(0,col_for_parent)
    else:
        rec_files = [(pathlib.Path(rec)/file_name,{'recording': rec.name}) for rec in rec_dirs]
        rec_files = [f for f in rec_files if f[0].is_file()]
    if not rec_files:
        return None, None, None
    df = pd.concat((pd.read_csv(rec[0], delimiter='\t').assign(**rec[1]) for rec in rec_files), ignore_index=True)
    if df.empty:
        return None, None, None
    # set indices
    df = df.set_index(idx_vals+['marker_interval','type','target'])
    # change type index into enum
    typeIdx = df.index.names.index('type')
    df.index = df.index.set_levels(pd.CategoricalIndex([getattr(DataQualityType,x) for x in df.index.levels[typeIdx]]),level='type')

    # see what we have
    dq_types = sorted(list(df.index.levels[typeIdx]), key=lambda dq: dq.value)
    targets  = list(df.index.levels[df.index.names.index('target')])

    # good default selection of dq type to export
    if DataQualityType.pose_vidpos_ray in dq_types:
        default_dq_type = DataQualityType.pose_vidpos_ray
    elif DataQualityType.pose_vidpos_homography in dq_types:
        default_dq_type = DataQualityType.pose_vidpos_homography
    else:
        # ultimate fallback, just set first available as the one to export
        default_dq_type = dq_types[0]

    return df, default_dq_type, targets

def summarize_and_store_data_quality(df: pd.DataFrame, output_file_or_dir: str | pathlib.Path, dq_types: list[DataQualityType], targets: list[int], average_over_targets = False, include_data_loss = False):
    dq_types_have = sorted(list(df.index.levels[df.index.names.index('type')]), key=lambda dq: dq.value)
    targets_have  = list(df.index.levels[df.index.names.index('target')])

    # remove unwanted types of data quality
    dq_types_sel = [dq in dq_types for dq in dq_types_have]
    if not all(dq_types_sel):
        df = df.drop(index=[dq for i,dq in enumerate(dq_types_have) if not dq_types_sel[i]], level='type')
    # remove unwanted targets
    targets_sel = [t in targets for t in targets_have]
    if not all(targets_sel):
        df = df.drop(index=[t for i,t in enumerate(targets_have) if not targets_sel[i]], level='target')
    # remove unwanted data loss
    if not include_data_loss and 'data_loss' in df.columns:
        df = df.drop(columns='data_loss')
    # average data if wanted
    if average_over_targets:
        gb = df.drop(columns='order').groupby([n for n in df.index.names if n!='target'],observed=True)
        count = gb.count()
        df = gb.mean()
        # add number of targets count (there may be some missing data)
        df.insert(0,'num_targets',count['acc'])

    # store
    output_file_or_dir = pathlib.Path(output_file_or_dir)
    if output_file_or_dir.is_dir():
        output_file_or_dir = output_file_or_dir / 'dataQuality.tsv'
    df.to_csv(output_file_or_dir, mode='w', header=True, sep='\t', na_rep='nan', float_format="%.6f")

def export_data_quality(rec_dirs: list[str | pathlib.Path], output_file_or_dir: str | pathlib.Path, dq_types: list[DataQualityType] = None, targets: list[int] = None, average_over_targets = False, include_data_loss = False):
    df, default_dq_type, targets_have = collect_data_quality(rec_dirs)
    if not dq_types:
        dq_types = [default_dq_type]
    if not targets:
        targets = targets_have
    summarize_and_store_data_quality(df, output_file_or_dir, dq_types, targets, average_over_targets, include_data_loss)


import numpy as np
import cv2
import pathlib
import shutil
import os
import subprocess
from enum import Enum, auto

from ffpyplayer.writer import MediaWriter
from ffpyplayer.pic import Image
import ffpyplayer.tools
from fractions import Fraction

from . import gaze_headref, naming, ocv, timestamps, video_utils
from .gui import video_player

class Status(Enum):
    Ok = auto()
    Skip = auto()
    Finished = auto()

class VideoMaker:
    def __init__(self, output_path: str|pathlib.Path, video_file: str|pathlib.Path, frame_timestamp_file: str|pathlib.Path|timestamps.VideoTimestamps, gaze_data_file: str|pathlib.Path|dict[int,list[gaze_headref.Gaze]]):
        self.src_video   = pathlib.Path(video_file)
        self.output_path = pathlib.Path(output_path)
        if self.output_path.is_dir():
            self.output_path /= naming.gaze_overlay_video_file
        self.video_ts   = frame_timestamp_file if isinstance(frame_timestamp_file,timestamps.VideoTimestamps) else timestamps.VideoTimestamps(frame_timestamp_file)
        self.video      = ocv.CV2VideoReader(self.src_video, self.video_ts.timestamps)
        self.gaze       = gaze_data_file if isinstance(gaze_data_file,dict) else gaze_headref.read_dict_from_file(gaze_data_file)[0]

        self._cache: tuple[Status, tuple[np.ndarray, int, float]] = None  # self._cache[1][1] is frame number

        self.gui                    : video_player.GUI  = None
        self.has_gui                                    = False

        self.do_visualize                               = False
        self.sub_pixel_fac                              = 8

        self._first_frame                               = True
        self._do_report_frames                          = True

        # set up output video
        self._fps= 1000/self.video_ts.get_IFI()
        self._res= (int(self.video.get_prop(cv2.CAP_PROP_FRAME_HEIGHT)),int(self.video.get_prop(cv2.CAP_PROP_FRAME_WIDTH)))
        codec    = ffpyplayer.tools.get_format_codec(fmt=self.output_path.suffix[1:])
        pix_fmt  = ffpyplayer.tools.get_best_pix_fmt('bgr24',ffpyplayer.tools.get_supported_pixfmts(codec))
        fpsFrac  = Fraction(self._fps).limit_denominator(10000).as_integer_ratio()
        out_opts = {'pix_fmt_in':'bgr24', 'pix_fmt_out':pix_fmt, 'width_in':self._res[1], 'height_in':self._res[0], 'frame_rate':fpsFrac}
        self._vid_writer = MediaWriter(str(self.output_path), [out_opts], overwrite=True)

    def __del__(self):
        if self.has_gui:
            self.gui.stop()

    def attach_gui(self, gui: video_player.GUI, window_id: int = None):
        self.gui                    = gui
        self.has_gui                = self.gui is not None
        self.do_visualize           = self.has_gui

        if self.has_gui:
            self.gui.set_show_timeline(True, self.video_ts, None, window_id)

    def set_visualize_on_frame(self, do_visualize: bool):
        self.do_visualize           = do_visualize

    def set_do_report_frames(self, do_report_frames: bool):
        self._do_report_frames = do_report_frames

    def _process_one_frame_impl(self, wanted_frame_idx:int = None) -> tuple[Status, tuple[np.ndarray, int, float]]:
        if self._first_frame and self.has_gui:
            self.gui.set_playing(True)
            self.gui.set_frame_size(self._res)
            self._first_frame = False

        if wanted_frame_idx is not None and self._cache is not None and self._cache[1][1]==wanted_frame_idx:
            return self._cache

        should_exit, frame, frame_idx, frame_ts = self.video.read_frame(report_gap=True, wanted_frame_idx=wanted_frame_idx)

        if should_exit:
            self._cache = Status.Finished, (None, None, None)
            return self._cache
        if self._do_report_frames:
            self.video.report_frame()

        if self.has_gui:
            requests = self.gui.get_requests()
            for r,_ in requests:
                if r=='exit':   # only request we need to handle
                    self._cache = Status.Finished, (None, None, None)
                    return self._cache

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.set(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        if frame is None:
            # we don't have a valid frame or nothing to do, continue to next
            if self.has_gui:
                # do update timeline of the viewers
                self.gui.update_image(None, frame_ts/1000., frame_idx)
            self._cache = Status.Skip, (frame, frame_idx, frame_ts)
            return self._cache

        # now that all processing is done, handle gui
        if self.has_gui:
            self.gui.update_image(frame, frame_ts/1000., frame_idx)

        self._cache = Status.Ok, (frame, frame_idx, frame_ts)
        return self._cache

    def process_one_frame(self) -> Status:
        status, (frame, frame_idx, _) = self._process_one_frame_impl()
        if status==Status.Finished:
            return status
        if frame is None:
            # get black frame
            frame = np.zeros(self._res,'uint8')

        # draw gaze
        if frame_idx in self.gaze:
            for g in self.gaze[frame_idx]:
                g.draw(frame, sub_pixel_fac=self.sub_pixel_fac)

        # submit frame to be encoded
        img = Image(plane_buffers=[frame.flatten().tobytes()], pix_fmt='bgr24', size=(frame.shape[1], frame.shape[0]))
        self._vid_writer.write_frame(img=img, pts=frame_idx/self._fps)

    def finish_video(self):
        self._vid_writer.close()
        # if ffmpeg is on path, add audio to scene video
        if (shutil.which('ffmpeg') is not None) and (shutil.which('ffprobe') is not None):
            # check if source file has audio
            command = ['ffprobe',
               '-loglevel', 'error',
               '-select_streams', 'a',
               '-show_entries', 'stream=codec_type',
               '-of', 'csv=p=0',
               f'{self.src_video}']
            proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = proc.communicate()
            if not err and out.decode().strip()=='audio':
                # file has audio, lets go
                # move file to temp name
                tempName = self.output_path.with_stem(self.output_path.stem + '_temp')
                shutil.move(self.output_path, tempName)

                # add audio
                cmds = ['ffmpeg', '-hide_banner', '-loglevel', 'error', '-y', '-i', f'"{tempName}"', '-i', f'"{self.src_video}"', '-vcodec', 'copy']
                # determine if can copy audio or need to transcode
                # we transcode if source file is not mp4-style, to be safe. Some audio formats that are allowed in
                # avi and mkv are not allowed in mp4
                if video_utils.is_isobmmf(self.src_video):
                    # copy. else, specifying nothing, it will be transcoded
                    cmds.extend(['-acodec', 'copy'])
                cmds.extend(['-map', '0:v:0', '-map', '1:a:0?', '-shortest', f'"{self.output_path}"'])
                os.system(' '.join(cmds))

                # clean up
                if self.output_path.exists():
                    tempName.unlink(missing_ok=True)
                else:
                    # something failed. Put file without audio back under output name
                    shutil.move(tempName, self.output_path)

    def process_video(self):
        while True:
            status = self.process_one_frame()
            if status==Status.Finished:
                break
            if status==Status.Skip:
                continue

        self.finish_video()


import enum

from . import config
from .. import utils as _utils

# NB: using pose information requires a calibrated scene camera
class DataQualityType(enum.Enum):
    viewpos_vidpos_homography   = enum.auto()   # use homography to map gaze from video to poster, and viewing distance defined in config (combined with the assumptions that the viewing position (eye) is located directly in front of the poster's center and that the poster is oriented perpendicularly to the line of sight) to compute angular measures
    pose_vidpos_homography      = enum.auto()   # use homography to map gaze from video to poster, and pose information w.r.t. poster to compute angular measures
    pose_vidpos_ray             = enum.auto()   # use camera calibration to map gaze position on scene video to cyclopean gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_world_eye              = enum.auto()   # use provided gaze position in world (often a binocular gaze point), and pose information w.r.t. poster to compute angular measures
    pose_left_eye               = enum.auto()   # use provided left eye gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_right_eye              = enum.auto()   # use provided right eye gaze vector, and pose information w.r.t. poster to compute angular measures
    pose_left_right_avg         = enum.auto()   # report average of left (pose_left_eye) and right (pose_right_eye) eye angular values

    # so this gets serialized in a user-friendly way by pandas..
    def __str__(self):
        return self.name
_utils.register_type(_utils.CustomTypeEntry(DataQualityType, 'glassesValidator.DataQualityType', _utils.enum_val_2_str, lambda x: _utils.enum_str_2_val(x, DataQualityType)))

def get_DataQualityType_explanation(dq: DataQualityType):
    ler_name =  "Left eye ray + pose"
    rer_name = "Right eye ray + pose"
    match dq:
        case DataQualityType.viewpos_vidpos_homography:
            return "Homography + view distance", \
                   "Use homography to map gaze position from the scene video to " \
                   "the validation poster, and use an assumed viewing distance (see " \
                   "the project's configuration) to compute data quality measures " \
                   "in degrees with respect to the scene camera. In this mode, it is "\
                   "assumed that the eye is located exactly in front of the center of "\
                   "the poster and that the poster is oriented perpendicularly to the "\
                   "line of sight from this assumed viewing position."
        case DataQualityType.pose_vidpos_homography:
            return "Homography + pose", \
                   "Use homography to map gaze position from the scene video to " \
                   "the validation poster, and use the determined pose of the scene " \
                   "camera (requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_vidpos_ray:
            return "Video ray + pose", \
                   "Use camera calibration to turn gaze position from the scene " \
                   "video into a direction vector, and determine gaze position on " \
                   "the validation poster by intersecting this vector with the " \
                   "poster. Then, use the determined pose of the scene camera " \
                   "(requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_world_eye:
            return "World gaze position + pose", \
                   "Use the gaze position in the world provided by the eye tracker " \
                   "(often a binocular gaze point) to determine gaze position on the " \
                   "validation poster by turning it into a direction vector with respect " \
                   "to the scene camera and intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a calibrated camera) to compute data quality " \
                   "measures in degrees with respect to the scene camera."
        case DataQualityType.pose_left_eye:
            return ler_name, \
                   "Use the gaze direction vector for the left eye provided by " \
                   "the eye tracker to determine gaze position on the validation " \
                   "poster by intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a camera calibration) to compute data quality " \
                   "measures in degrees with respect to the left eye."
        case DataQualityType.pose_right_eye:
            return rer_name, \
                   "Use the gaze direction vector for the right eye provided by " \
                   "the eye tracker to determine gaze position on the validation " \
                   "poster by intersecting this vector with the poster. " \
                   "Then, use the determined pose of the scene camera " \
                   "(requires a camera calibration) to compute data quality " \
                   "measures in degrees with respect to the right eye."
        case DataQualityType.pose_left_right_avg:
            return "Average eye rays + pose", \
                   "For each time point, take angular offset between the left and " \
                   "right gaze positions and the fixation target and average them " \
                   "to compute data quality measures in degrees. Requires " \
                   f"'{ler_name}' and '{rer_name}' to be enabled."